#include <pcl/io/io.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/surface/mls.h>
#include <pcl/common/transforms.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <boost/thread/thread.hpp>
#include <pcl/visualization/common/shapes.h>
#include <pcl/filters/passthrough.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/features/normal_3d_omp.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/radius_outlier_removal.h>
#include <iostream>
#include <fstream>
#include <chrono>
#include <thread>
#include <algorithm>
#include "./utility.cpp"


typedef enum{
	PREWITT = 0,
	SOBEL,
	SCHARR,
	LAPLACIAN,
} detect_mask_t;

typedef struct {
	float gradient = 0;
	float laplacian = 0; 
	int angle = 0;
	Eigen::Matrix3i neighbors;
} edge_t_e;

typedef struct {
	float p;
	float n;
} ext_edge_t;

Eigen::Matrix3i get_neighbors(pcl::PointXY center, int center_index, float distance_between_points, pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, std::vector<int> neighbors)
{
	Eigen::Matrix3i m;
	m(0,0) = center_index;
	m(0,1) = center_index;
	m(0,2) = center_index;
	m(1,0) = center_index;
	m(1,1) = center_index;
	m(1,2) = center_index;
	m(2,0) = center_index;
	m(2,1) = center_index;
	m(2,2) = center_index;

	float threshold = distance_between_points*0.25;
	int iter = 0;
	for (auto &p : neighbors)
	{
		float nx = cloud->points[p].x;
		float ny = cloud->points[p].y;

		if(p == center_index or iter > 8) 
		{
			continue;
		}

		if ( nx > (center.x + threshold) and ny > (center.y + threshold) and (m(0,0) == center_index))
		{
			m(0,0) = p;
		} 
		else if ( nx < (center.x - threshold) and ny > (center.y + threshold) and (m(0,2) == center_index))
		{
			m(0,2) = p;
		}
		else if ( nx > (center.x + threshold) and ny < (center.y - threshold) and (m(2,0) == center_index))
		{
			m(2,0) = p;
		}
		else if ( nx < (center.x - threshold) and ny < (center.y - threshold) and (m(2,2) == center_index))
		{
			m(2,2) = p;
		}
		else if ( nx > (center.x - threshold) and nx < (center.x + threshold) and ny > (center.y + threshold) and (m(0,1) == center_index))
		{
			m(0,1) = p;
		}
		else if ( nx > (center.x - threshold) and nx < (center.x + threshold) and ny < (center.y - threshold) and (m(2,1) == center_index))
		{
			m(2,1) = p;
		}
		else if ( ny > (center.y - threshold) and ny < (center.y + threshold) and nx > (center.x + threshold) and (m(1,0) == center_index))
		{
			m(1,0) = p;
		}
		else if ( ny > (center.y - threshold) and ny < (center.y + threshold) and nx < (center.x - threshold) and (m(1,2) == center_index))
		{
			m(1,2) = p;
		}
		iter++;
	}

	return m;
}

int normalized_angle(float x, float y)
{
	float angle = atan2(y, x);

	//std::cout << "Angle: " << angle << std::endl;
 
	if (angle < M_PI/4 and angle > -M_PI/4)
	{
		return 1;
	}
	else if (angle < -(3*M_PI/4) or angle > (3*M_PI/4))
	{
		return -1;
	}
	else return 0;

	/*
	if ( angle > 0)
	{
		if (angle < (M_PI/8))
		{
			angle = 0;
		}
		else if (angle > (M_PI/8) and angle < (3*M_PI/8))
		{
			angle = 1;
		}
		else if (angle > (3*M_PI/8) and angle < (5*M_PI/8))
		{
			angle = 2;
		}
		else if (angle > (5*M_PI/8) and angle < (7*M_PI/8))
		{
			angle = 3;
		}
	} else if ( angle < 0)
	{
		if (angle > -(M_PI/8))
		{
			angle = 7;
		}
		else if (angle < -(M_PI/8) and angle > -(3*M_PI/8))
		{
			angle = 6;
		}
		else if (angle < -(3*M_PI/8) and angle > -(5*M_PI/8))
		{
			angle = 5;
		}
		else if (angle < -(5*M_PI/8) and angle > -(7*M_PI/8))
		{
			angle = 4;
		}
	*/
}



typedef struct 
{
	bool lpmin = false;
	bool lpmax = false;
	bool lpzc = false;
	bool gmin = false;
	bool gmax = false;
	bool edge_neighbor = false;
	bool edge = false;
}zc_classifier_t;


std::vector<bool> find_zero_crossing(std::vector<edge_t_e> gradients, pcl::PointCloud<pcl::PointXYZRGB>::Ptr input, float length, float stepsize)
{
	std::vector<zc_classifier_t> classifiers;

	for (int i = 0; i < input->size(); ++i)
	{
		zc_classifier_t local_classifier;

		auto n = gradients[i].neighbors;

		float pos_x_lapl = gradients[n(1,0)].laplacian;
		float neg_x_lapl = gradients[n(1,2)].laplacian;

		float pos_x_grad = gradients[n(1,0)].gradient;
		float neg_x_grad = gradients[n(1,2)].gradient;

		float cur_lapl = gradients[i].laplacian;
		float cur_grad = gradients[i].gradient;

		if((neg_x_lapl > 0 and cur_lapl < 0) or (cur_lapl < 0 and pos_x_lapl < 0) or (neg_x_lapl < 0 and cur_lapl > 0) or (cur_lapl > 0 and pos_x_lapl < 0))
		{
			//(neg_x_lapl > 0 and pos_x_lapl < 0 and neg_x_lapl > cur_lapl and pos_x_lapl < cur_lapl ) or (neg_x_lapl < 0 and pos_x_lapl > 0 and neg_x_lapl < cur_lapl and pos_x_lapl > cur_lapl ) or (neg_x_lapl > 0 and cur_lapl < 0 and pos_x_lapl > 0) or (neg_x_lapl < 0 and cur_lapl > 0 and pos_x_lapl < 0)
			local_classifier.lpzc = true;
		}

		if (cur_lapl < pos_x_lapl and cur_lapl < neg_x_lapl)
		{
			local_classifier.lpmin = true;
		}

		if (cur_lapl > pos_x_lapl and cur_lapl > neg_x_lapl)
		{
			local_classifier.lpmin = true;
		}
		
		if (cur_grad > pos_x_grad and cur_grad > neg_x_grad)
		{
			local_classifier.gmax = true;
		} 
		else if (cur_grad < pos_x_grad and cur_grad < neg_x_grad)
		{
			local_classifier.gmin = true;
		}

		classifiers.push_back(local_classifier);
	}
	for (int i = 1; i < input->size(); ++i)
	{

		auto n = gradients[i].neighbors;
		
		auto pos_lc = classifiers[n(1,0)];
		auto neg_lc = classifiers[n(1,2)];
		auto cur_lc = classifiers[i];

		if (cur_lc.gmin and (cur_lc.lpzc or neg_lc.lpzc or pos_lc.lpzc))
		{
			classifiers[i].edge = true;
			classifiers[i+1].edge_neighbor = true;
			classifiers[i-1].edge_neighbor = true;
		}
	}
	for (int i = 0; i < input->size(); ++i)
	{
		auto n = gradients[i].neighbors;
		
		auto pos_lc = classifiers[n(1,0)];
		auto neg_lc = classifiers[n(1,2)];
		auto cur_lc = classifiers[i];

		if (gradients[i].angle < 0)
		{
			if (cur_lc.gmax and (cur_lc.lpzc or pos_lc.lpzc) and !(neg_lc.gmin or pos_lc.gmin))
			{
				if (!(neg_lc.edge or pos_lc.edge or pos_lc.edge_neighbor or neg_lc.edge_neighbor))
				{
					classifiers[i].edge = true;
				}
			}
		} else if (gradients[i].angle > 0)
		{
			if (cur_lc.gmax and (cur_lc.lpzc or neg_lc.lpzc) and !(neg_lc.gmin or pos_lc.gmin))
			{
				if (!(neg_lc.edge or pos_lc.edge or pos_lc.edge_neighbor or neg_lc.edge_neighbor))
				{
					classifiers[i].edge = true;
				}
			}
		}	
	}
	std::vector<bool> edges;

	for (int i = 1; i < input->size(); ++i)
	{
		if (classifiers[i].edge)
		{
			edges.push_back(true);

			input->points[i].r = 255;
			input->points[i].g = 255;
			input->points[i].b = 255;	
		} 
		else 
		{
			edges.push_back(false);
			input->points[i].r = 0;
			input->points[i].g = 0;
			input->points[i].b = 0;
		}
	}
	return edges;
}

typedef struct{
	float gmax = 0.0;
	float gmin = 0.0;
	float neighbor_gmax = 0.0;
	bool edge = false;
} grad_classifier_t;

std::vector<bool> find_edges_from_gradient(std::vector<edge_t_e> gradients, pcl::PointCloud<pcl::PointXYZRGB>::Ptr input, float start_y, float end_y)
{
	std::vector<grad_classifier_t> classifiers;

	std::vector<bool> edges;

	for (int i = 0; i < input->size(); ++i)
	{
		/*grad_classifier_t local_classifier;*/

		if (input->points[i].y > end_y or input->points[i].y < start_y or std::abs(input->points[i].x) > 0.01)
		{
			//classifiers.push_back(local_classifier);
			edges.push_back(false);
			continue;
		}

		auto n = gradients[i].neighbors;

		float pos_x_grad = gradients[n(1,0)].gradient;
		float neg_x_grad = gradients[n(1,2)].gradient;
		float cur_grad = gradients[i].gradient;

		if (cur_grad > pos_x_grad and cur_grad > neg_x_grad)
		{
			edges.push_back(true);
			// local_classifier.gmax = cur_grad;
		} else edges.push_back(false);
		// else if (cur_grad < pos_x_grad and cur_grad < neg_x_grad)
		// 		{
		// 			local_classifier.gmin = cur_grad;
		// 		}

		// 		//classifiers.push_back(local_classifier);
		// 	}

		// 	for (int i = 0; i < input->size(); ++i)
		// 	{
		// 		if (input->points[i].y > end_y or input->points[i].y < start_y or std::abs(input->points[i].x) > 0.01)
		// 		{
		// 			continue;
		// 		}

		// 		auto n = gradients[i].neighbors;

		// 		if ((classifiers[n(1,2)].neighbor_gmax < classifiers[i].gmax) and (classifiers[n(1,0)].neighbor_gmax < classifiers[i].gmax))
		// 		{
		// 			classifiers[i].edge = true;
		// 			classifiers[n(1,2)].neighbor_gmax = classifiers[i].gmax;
		// 			classifiers[n(1,0)].neighbor_gmax = classifiers[i].gmax;	
		// 		} 
		// 	}

		// 	for (int i = 0; i < input->size(); ++i)
		// 	{
		// 		if (input->points[i].y > end_y or input->points[i].y < start_y or std::abs(input->points[i].x) > 0.01)
		// 		{
		// 			continue;
		// 		}

		// 		auto n = gradients[i].neighbors;

		// 		if ((classifiers[n(1,2)].neighbor_gmax < classifiers[i].gmax) and (classifiers[n(1,0)].neighbor_gmax < classifiers[i].gmax))
		// 		{
		// 			classifiers[i].edge = true;
		// 			classifiers[n(1,2)].neighbor_gmax = classifiers[i].gmax;
		// 			classifiers[n(1,0)].neighbor_gmax = classifiers[i].gmax;	
		// 		} else {
		// 			classifiers[i].edge = false;
		// 		}
	}
	return edges;
}


std::vector<ext_edge_t> road_width_from_edges(std::vector<bool> all_edges, pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, float length, float stepsize)
{
	float sum_pos_edges = 0.0;
	int num_pos_edges = 0;
	float sum_neg_edges = 0.0;
	int num_neg_edges = 0;

	std::vector<ext_edge_t> edges;

	//std::cout << "Cloud size: " << cloud->size() << std::endl;  
	for (float step = 0; step < length; step += stepsize)
	{	
		ext_edge_t edge;

		for (int i = 0; i < cloud->size(); ++i)
		{
			if (cloud->points[i].y > 0.0025+step or cloud->points[i].y < step or std::abs(cloud->points[i].x) > 0.01)
			{
				continue;
			}


			float x = cloud->points[i].x;
			float y = cloud->points[i].y;
			float z = cloud->points[i].z;


			if (all_edges[i] and x > 0)
			{
				sum_pos_edges += x;
				num_pos_edges++;
			} 
			else if (all_edges[i] and x < 0)
			{
				sum_neg_edges += x;
				num_neg_edges ++;
			}

		}
		edge.p = sum_pos_edges/static_cast<float>(num_pos_edges);
		edge.n = sum_neg_edges/static_cast<float>(num_neg_edges);
		edges.push_back(edge);
	}

	return edges;
}




ext_edge_t road_width_by_gradient(std::vector<edge_t_e> gradients, pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, float start_y, float end_y, float lower_z)
{
	float grad_p[200] = {0};
	float lapl_p[200] = {0};
	int num_p[200] = {0};

	float grad_n[200] = {0};
	float lapl_n[200] = {0};
	int num_n[200] = {0};

	std::vector<ext_edge_t> edges;

	//std::cout << "Cloud size: " << cloud->size() << std::endl;  
	ext_edge_t edge;

	for (int i = 0; i < cloud->size(); ++i)
	{
		if (cloud->points[i].y > end_y or cloud->points[i].y < start_y or std::abs(cloud->points[i].x) > 0.01 or cloud->points[i].z < lower_z)
		{
			continue;
		}
		

		for  (int dx = 0; dx < 150; ++dx)
		{
			float decimal_x = static_cast<float>(dx)/20000;
			float rounded_x = round(cloud->points[i].x,20000);
			//std::cout << rounded_x << std::endl;

			//std::cout << gradients[i].angle << std::endl;

			if ((rounded_x >= decimal_x - 0.000025) and (rounded_x < decimal_x + 0.000025))
			{
				if (gradients[i].angle < 0)
				{
					grad_p[dx] += gradients[i].gradient;
					//lapl_p[dx] += gradients[i].laplacian;

					num_p[dx]++;	/* code */
				}
				continue;	
			}

			if ((rounded_x <= -decimal_x + 0.000025) and (rounded_x > -decimal_x - 0.000025))
			{
				if (gradients[i].angle > 0)
				{
					
					
					grad_n[dx] += gradients[i].gradient;
					//lapl_n[dx] += gradients[i].laplacian;
					num_n[dx]++;
					//std::cout << "Negative side of 0" << std::endl;
				}
				continue;
			}
		}
	}

	float biggest_grad_p = 0.0;
	int biggest_grad_p_index = 0;
	float biggest_grad_n = 0.0;
	int biggest_grad_n_index = 0;

	for (int i = 0; i < 100; ++i)
	{
		if (num_p[i] > 0 and grad_p[i] > 0)
		{

			//std::cout << "Total Gradient at " << static_cast<float>(i)/10 << "mm is " << grad_p[i] << ", with number of points " << num_p[i]; 
			grad_p[i] = grad_p[i]/static_cast<float>(num_p[i]);
			//std::cout << " with average gradient " << grad_p[i] << std::endl;

			if (grad_p[i] > biggest_grad_p)
			{
				biggest_grad_p = grad_p[i];
				biggest_grad_p_index = i;
			} else if (grad_p[i] == biggest_grad_p){
				biggest_grad_p = grad_p[i];
				biggest_grad_p_index = i;
			}
		}
		if (num_n[i] > 0 and grad_n[i] > 0)
		{
			//std::cout << "Total Gradient at " << -static_cast<float>(i)/10 << "mm is " << grad_n[i] << ", with number of points " << num_n[i]; 
			grad_n[i] = grad_n[i]/static_cast<float>(num_n[i]);
			//std::cout << ", with average gradient " << grad_n[i] << std::endl;
			
			if (grad_n[i] > biggest_grad_n)
			{
				biggest_grad_n = grad_n[i];
				biggest_grad_n_index = i;
			} else if (grad_n[i] == biggest_grad_n){
				biggest_grad_n = grad_n[i];
				biggest_grad_n_index = i;
			}
		}
	}

	//std::cout << "Highest positive gradient: " << biggest_grad_p << std::endl;
	//std::cout << "Highest negative gradient: " << biggest_grad_n << std::endl;

	float edge_p = static_cast<float>(biggest_grad_p_index)/20000;
	float edge_n = -static_cast<float>(biggest_grad_n_index)/20000; 

	//std::cout << "Positive Edge: " << edge_p << " Negative Edge: " << edge_n << std::endl;

	edge.p = edge_p;
	edge.n = edge_n;

	return edge;
}

float cont_gaussian_blur(float dist, float sigma)
{
	float exponent = -(dist*dist)/(2*sigma*sigma);
	float result = exp(exponent);
	return result;
}

void gaussian_blur(pcl::PointCloud<pcl::PointXYZRGB>::Ptr input, pcl::PointCloud<pcl::PointXYZRGB>::Ptr output, int size, float point_distance)
{
	Eigen::Matrix3f gaussian;
  	gaussian(0,0) = 1.0;
  	gaussian(1,0) = 2.0;
  	gaussian(2,0) = 1.0;
  	gaussian(0,1) = 2.0;
  	gaussian(1,1) = 4.0;
  	gaussian(2,1) = 2.0;
  	gaussian(0,2) = 1.0;
  	gaussian(1,2) = 2.0;
  	gaussian(2,2) = 1.0;

  	pcl::copyPointCloud(*input, *output);

  	pcl::PointCloud<pcl::PointXY>::Ptr xy_cloud (new pcl::PointCloud<pcl::PointXY>);
  	pcl::copyPointCloud(*input,*xy_cloud);
	pcl::KdTreeFLANN<pcl::PointXY> kdtree;
	kdtree.setInputCloud(xy_cloud);
	
	float sigma = (size*point_distance)/6;

	std::cout << "sigma: " << sigma << std::endl;

  	pcl::PointXY searchPoint;

  	int K = size*size;

  	for (int i = 0; i < input->size(); ++i)
  	{
  		searchPoint.x = input->points[i].x;
  		searchPoint.y = input->points[i].y;
  		std::vector<int> pointIdxNKNSearch(K);
  		std::vector<float> pointNKNSquaredDistance(K);
  		if ( !(kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0 ))
  		{
  			continue;
  		}
  		float blurred_z = 0.0;
  		for (int i = 0; i < K; ++i)
  		{
  			blurred_z += input->points[pointIdxNKNSearch[i]].z*cont_gaussian_blur(pointNKNSquaredDistance[i],sigma); 
  		}
  		/*auto n = get_neighbors(searchPoint, i, point_distance, input, pqqqqqqqqqqqointIdxNKNSearch);

		for (int q = 0; q < 3; q++)
		{
			//std::cout << "| ";
			for (int p = 0; p < 3; p++)
			{  
				blurred_z += gaussian(q,p) * input->points[n(q,p)].z;
			}
		} 

		output->points[i].z = blurred_z/16.0;	*/	
		output->points[i].z = blurred_z/K;
  	}
}

float LoG(float dist,float sigma)
{
	float exponent = (dist*dist)/(2*sigma*sigma);
	float factor = ((dist*dist)- 2*sigma*sigma)/(sigma*sigma*sigma*sigma);
	float result = factor * exp(-exponent);
	return result;
}


std::vector<edge_t_e> calc_gradient(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, int gradient_type)
{	

	pcl::PointCloud<pcl::PointXY>::Ptr xy_cloud (new pcl::PointCloud<pcl::PointXY>);
	pcl::copyPointCloud(*cloud,*xy_cloud);

	pcl::KdTreeFLANN<pcl::PointXY> kdtree;

  	kdtree.setInputCloud(xy_cloud);

  	pcl::PointXY searchPoint;
  	std::vector<edge_t_e> point_intensities;

  	Eigen::Matrix3f prewitt_v;
  	prewitt_v(0,0) = 1.0;
  	prewitt_v(1,0) = 1.0;
  	prewitt_v(2,0) = 1.0;
  	prewitt_v(0,1) = 0.0;
  	prewitt_v(1,1) = 0.0;
  	prewitt_v(2,1) = 0.0;
  	prewitt_v(0,2) = -1.0;
  	prewitt_v(1,2) = -1.0;
  	prewitt_v(2,2) = -1.0;

  	Eigen::Matrix3f prewitt_h;
  	prewitt_h(0,0) = 1.0;
  	prewitt_h(0,1) = 1.0;
  	prewitt_h(0,2) = 1.0;
  	prewitt_h(1,0) = 0.0;
  	prewitt_h(1,1) = 0.0;
  	prewitt_h(1,2) = 0.0;
  	prewitt_h(2,0) = -1.0;
  	prewitt_h(2,1) = -1.0;
  	prewitt_h(2,2) = -1.0;
	
  	Eigen::Matrix3f sobel_v;
  	sobel_v(0,0) = 1.0;
  	sobel_v(1,0) = 2.0;
  	sobel_v(2,0) = 1.0;
  	sobel_v(0,1) = 0.0;
  	sobel_v(1,1) = 0.0;
  	sobel_v(2,1) = 0.0;
  	sobel_v(0,2) = -1.0;
  	sobel_v(1,2) = -2.0;
  	sobel_v(2,2) = -1.0;

  	Eigen::Matrix3f sobel_h;
  	sobel_h(0,0) = 1.0;
  	sobel_h(0,1) = 2.0;
  	sobel_h(0,2) = 1.0;
  	sobel_h(1,0) = 0.0;
	sobel_h(1,1) = 0.0;
	sobel_h(1,2) = 0.0;
  	sobel_h(2,0) = -1.0;
  	sobel_h(2,1) = -2.0;
  	sobel_h(2,2) = -1.0;

  	Eigen::Matrix3f scharr_v;
  	scharr_v(0,0) = 3.0;
  	scharr_v(1,0) = 10.0;
  	scharr_v(2,0) = 3.0;
  	scharr_v(0,1) = 0.0;
  	scharr_v(1,1) = 0.0;
  	scharr_v(2,1) = 0.0;
  	scharr_v(0,2) = -3.0;
  	scharr_v(1,2) = -10.0;
  	scharr_v(2,2) = -3.0;

  	Eigen::Matrix3f scharr_h;
  	scharr_h(0,0) = 3.0;
  	scharr_h(0,1) = 10.0;
  	scharr_h(0,2) = 3.0;
  	scharr_h(1,0) = 0.0;
	scharr_h(1,1) = 0.0;
	scharr_h(1,2) = 0.0;
  	scharr_h(2,0) = -3.0;
  	scharr_h(2,1) = -10.0;
  	scharr_h(2,2) = -3.0;

  	float point_distance = 0.0005;

	for (int i = 0; i < cloud->size(); ++i)
	{
		
		float point_threshold = 0.0001;

		searchPoint.x = cloud->points[i].x;
		searchPoint.y = cloud->points[i].y;

		float x = cloud->points[i].x;
		float y = cloud->points[i].y;
		float z = cloud->points[i].z;

		//Find nearest neighbors
		

  		float horizontal_gradient = 0.0;
  		float vertical_gradient = 0.0;
  		float laplacian = 0.0;

  		

  		int K = 25;	

  		std::vector<int> pointIdxNKNSearch(K);
  		std::vector<float> pointNKNSquaredDistance(K);


  		switch (gradient_type){
  			case 0:
			{
				edge_t_e current_point;

				if ( !(kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0 ))
		  		{
		  			continue;
		  		}

  				current_point.neighbors = get_neighbors(searchPoint, i, point_distance, cloud, pointIdxNKNSearch);
  				//current_point.neighbors(1,1) = i;

  				//std::cout << "Point: " << i << std::endl;

  				for (int q = 0; q < 3; q++)
  				{
  					//std::cout << "| ";
  					for (int p = 0; p < 3; p++)
  					{  
  						//std::cout << current_point.neighbors(q,p) << " ";
						vertical_gradient += prewitt_v(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  						horizontal_gradient += prewitt_h(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  					}
  					//std::cout << "|" << std::endl;
  				}

  				current_point.gradient = sqrt(horizontal_gradient*horizontal_gradient + vertical_gradient*vertical_gradient);
  				//std::cout << ">> gradient: " << current_point.gradient << ", h grad: " << horizontal_gradient << ", v grad: " << vertical_gradient << std::endl;
  				current_point.angle = normalized_angle(vertical_gradient, horizontal_gradient);

  				point_intensities.push_back(current_point);
  				break;
			}
  			case 1:
			{
				edge_t_e current_point;

				if ( !(kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0 ))
		  		{
		  			continue;
		  		}

  				current_point.neighbors = get_neighbors(searchPoint, i, point_distance, cloud, pointIdxNKNSearch);
  				//current_point.neighbors(1,1) = i;

  				//std::cout << "Point: " << i << std::endl;

  				for (int q = 0; q < 3; q++)
  				{
  					//std::cout << "| ";
  					for (int p = 0; p < 3; p++)
  					{  
  						//std::cout << current_point.neighbors(q,p) << " ";
						vertical_gradient += sobel_v(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  						horizontal_gradient += sobel_h(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  					}
  					//std::cout << "|" << std::endl;
  				}

  				horizontal_gradient = 0.0;

  				current_point.gradient = sqrt(horizontal_gradient*horizontal_gradient + vertical_gradient*vertical_gradient);
  				//std::cout << ">> gradient: " << current_point.gradient << ", h grad: " << horizontal_gradient << ", v grad: " << vertical_gradient << std::endl;
  				current_point.angle = normalized_angle(vertical_gradient, horizontal_gradient);

  				point_intensities.push_back(current_point);
  				break;
  			}
  			case 2:
			{
				edge_t_e current_point;

				if ( !(kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0 ))
		  		{
		  			continue;
		  		}

  				current_point.neighbors = get_neighbors(searchPoint, i, point_distance, cloud, pointIdxNKNSearch);
  				//current_point.neighbors(1,1) = i;

  				//std::cout << "Point: " << i << std::endl;

  				for (int q = 0; q < 3; q++)
  				{
  					//std::cout << "| ";
  					for (int p = 0; p < 3; p++)
  					{  
  						//std::cout << current_point.neighbors(q,p) << " ";
						vertical_gradient += scharr_v(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  						horizontal_gradient += scharr_h(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  					}
  					//std::cout << "|" << std::endl;
  				}

  				horizontal_gradient = 0.0;

  				current_point.gradient = sqrt(horizontal_gradient*horizontal_gradient + vertical_gradient*vertical_gradient);
  				current_point.angle = normalized_angle(vertical_gradient, horizontal_gradient);



  				point_intensities.push_back(current_point);
  				break;
  			}
  			case 3:
			{	
				edge_t_e current_point;

				if ( !(kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0 ))
		  		{
		  			continue;
		  		}

  				current_point.neighbors = get_neighbors(searchPoint, i, point_distance, cloud, pointIdxNKNSearch);
  				//current_point.neighbors(1,1) = i;

  				//std::cout << "Point: " << i << std::endl;

  				for (int q = 0; q < 3; q++)
  				{
  					//std::cout << "| ";
  					for (int p = 0; p < 3; p++)
  					{  
  						//std::cout << current_point.neighbors(q,p) << " ";
						vertical_gradient += scharr_v(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  						horizontal_gradient += scharr_h(q,p) * cloud->points[current_point.neighbors(q,p)].z;
  					}
  					//std::cout << "|" << std::endl;
  				}

  				//horizontal_gradient = 0.0;
  				//vertical_gradient = 0.0;

  				current_point.gradient = sqrt(horizontal_gradient*horizontal_gradient + vertical_gradient*vertical_gradient);
  				if (current_point.gradient < 0.0001)
  				{
  					current_point.gradient = 0;
  				}


  				current_point.angle = normalized_angle(vertical_gradient, horizontal_gradient);
  				//std::cout << "Angle: " << current_point.angle << std::endl;

  				int iter = 0;
  				for(auto &p : pointIdxNKNSearch)
  				{
  					// float dist = pointNKNSquaredDistance[iter];
  					// float log = LoG(dist,0.0005);
  					float log = 0.0;
  					if(p == i) log = 4; //16
  					else if (iter < 5) log = -1; //-2
  					else if (iter < 13) log = 0; //-1
  					else log = 0;


  					float nz = cloud->points[p].z*log;
  					//std::cout << "dist: " << dist << ", log: " << log << "height: " << cloud->points[p].z << ", nz: " << nz << std::endl;

  					laplacian += nz;
  					iter++;
  				}

  				if (std::abs(laplacian) < 0.0001)
  				{
  					laplacian = 0.0;
  				}

		  		//current_point.gradient = 0.0; //laplacian;
		  		current_point.laplacian = laplacian;

				if ( x > 0.0)
		  		{
		  			current_point.angle = -1;
		  		} else if (x < 0.0) 
		  		{
		  			current_point.angle = 1;
		  		}

		  		if (i > 300 and i < 360)
		  		{
		  			std::cout << current_point.gradient << "\t" << 10*current_point.laplacian << std::endl;
		  		}
  				
  				point_intensities.push_back(current_point);

  				break;
			}	
  		}
  	}

	/*for (auto &i : point_intensities)
 	{
 		int pos_neigbhour = 0;
 		int neg_neigbhour = 0;

 		switch (i.angle)
 		{
 			case 0:
 				pos_neigbhour = i.neighbors(1,2);
 				neg_neigbhour = i.neighbors(1,0);
 			case 1:
 				pos_neigbhour = i.neighbors(0,2);
 				neg_neigbhour = i.neighbors(2,0);
 			case 2:
 				pos_neigbhour = i.neighbors(0,1);
 				neg_neigbhour = i.neighbors(2,1);
 			case 3:
 				pos_neigbhour = i.neighbors(0,0);
 				neg_neigbhour = i.neighbors(2,2);
 			case 4:
 				pos_neigbhour = i.neighbors(1,0);
 				neg_neigbhour = i.neighbors(1,2);
 			case 5:
 				pos_neigbhour = i.neighbors(2,0);
 				neg_neigbhour = i.neighbors(0,2);
 			case 6:
 				pos_neigbhour = i.neighbors(2,1);
 				neg_neigbhour = i.neighbors(0,1);
 			case 7:
 				pos_neigbhour = i.neighbors(2,2);
 				neg_neigbhour = i.neighbors(0,0);
 		}

 		if (!(i.gradient > point_intensities[pos_neigbhour].gradient and i.gradient > point_intensities[neg_neigbhour].gradient))
 		{
 			i.gradient = 0.0;
 		}
 	}*/

  	float largest_gradient = 0.0;
  	float largest_laplacian = 0.0;

 	for (auto &i : point_intensities)
 	{
 		if (i.gradient > largest_gradient) largest_gradient = i.gradient;
 		if (i.laplacian > largest_laplacian) largest_laplacian = i.laplacian;
 		//std::cout << "gradient: " << i.gradient << std::endl;
 		
 	}

 	std::cout << "largest_gradient: " << largest_gradient << std::endl;

 	for (int i = 0; i < cloud->size(); ++i)
 	{

 		auto c = color_gradient_hot(point_intensities[i].gradient, largest_gradient); 

 		cloud->points[i].r = c[0];
 		cloud->points[i].g = c[1];
 		cloud->points[i].b = c[2];
 	}

 	return	point_intensities;
}