#include <pcl/io/io.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/surface/mls.h>
#include <pcl/common/transforms.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <boost/thread/thread.hpp>
#include <pcl/visualization/common/shapes.h>
#include <pcl/filters/passthrough.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/features/normal_3d_omp.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/radius_outlier_removal.h>
#include <iostream>
#include <fstream>
#include <chrono>
#include <thread>
#include <algorithm>
#include "./utility.cpp"
#include "./edge_detectors.cpp"
#include "./ideal_cloud.cpp"

pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);

Eigen::Matrix4f camera_transform()
{
	Eigen::Matrix4f camera_transform = Eigen::Matrix4f::Identity();
	
	float z_d = 0.108 + 0.092;
	float y_d = 0.101;
	float x_d = -0.004;//0466;//  ir_camera_centered: 0.00466;

	camera_transform (1,1) = -1;
	camera_transform (1,2) = 0.0f;
	camera_transform (2,1) = 0.0f;
	camera_transform (2,2) = -1;
	camera_transform (0,3) = x_d;
	camera_transform (1,3) = y_d;
	camera_transform (2,3) = z_d;

	return camera_transform;
}

Eigen::Matrix4f transform_from_plane_normal(pcl::ModelCoefficients::Ptr coeffs)
{
	double x = coeffs->values[0];
	double y = coeffs->values[1];
	double z = coeffs->values[2];
	double d = coeffs->values[3];

	std::cout << "Plane Coefficents" << std::endl;
	std::cout << "x: " << x << " y: " << y << " z: " << z << " d: " << d << std::endl;
	double s_a = y / sqrt(x*x + y*y);
	double c_a = x / sqrt(x*x + y*y);
	double s_b = sqrt(x*x + y*y);
	double c_b = z;

	Eigen::Matrix4f rot_z = Eigen::Matrix4f::Identity();
	rot_z (0,0) = c_a;
	rot_z (0,1) = -s_a;
	rot_z (1,0) = s_a;
	rot_z (1,1) = c_a;

	Eigen::Matrix4f rot_y = Eigen::Matrix4f::Identity();
	rot_y (0,0) = c_b;
	rot_y (0,2) = s_b;
	rot_y (2,0) = -s_b;
	rot_y (2,2) = c_b;

	Eigen::Matrix4f result = rot_z * rot_y;
	result (2,3) = -d;

	return rot_z*result.inverse();
}

Eigen::Matrix4f calibrate_planes(pcl::PointCloud<pcl::PointXYZ>::Ptr source)
{
	
	pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
	pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);

	pcl::SACSegmentation<pcl::PointXYZ> seg;
	seg.setOptimizeCoefficients(true);
	seg.setModelType(pcl::SACMODEL_PLANE);
	seg.setDistanceThreshold(0.00075);
	seg.setInputCloud(source);

	seg.segment(*inliers,*coefficients);
	if (inliers->indices.size () == 0)
	{
	   	PCL_ERROR ("Could not estimate a planar model for the given dataset.");
	}

	return transform_from_plane_normal(coefficients);
}

void range_filter_x_y(pcl::PointCloud<pcl::PointXYZ>::Ptr input, pcl::PointCloud<pcl::PointXYZ>::Ptr output, double min, double max)
{
	std::vector<int> indices;

	for (int i = 0; i < input->size(); ++i)
	{
		double x = input->points[i].x;
		double y = input->points[i].y;

		double length = sqrt(x*x + y*y);

		if ( length < max and length > min)
		{
			indices.push_back(i);
		}
	}

	pcl::copyPointCloud(*input, indices, *output);
}

void range_filter_x_y_z(pcl::PointCloud<pcl::PointXYZ>::Ptr input, pcl::PointCloud<pcl::PointXYZ>::Ptr output, pcl::PointCloud<pcl::PointXYZ>::Ptr center, int center_indice, double min, double max)
{
	std::vector<int> indices;
	float cx = center->points[center_indice].x;
	float cy = center->points[center_indice].y;
	float cz = center->points[center_indice].z;

	//std::cout << "x: " << center->points[center_indice].x << " y: " << center->points[center_indice].y << " z: " << center->points[center_indice].z << std::endl;

	for (int i = 0; i < input->size(); ++i)
	{	
		float dx = input->points[i].x - cx;
		float dy = input->points[i].y - cy;
		float dz = input->points[i].z - cz;

		float length = sqrt(dx*dx + dy*dy + dz*dz);

		if ( length < max and length > min)
		{
			indices.push_back(i);
 		}
	}

	//std::cout << "Size of input: " << input->size() << ", Filtered indices: " << indices.size();
	pcl::copyPointCloud(*input, indices, *output);
	//std::cout << " Size of output: " << output->size() << std::endl;
}

void range_filter_box_x_y_z(pcl::PointCloud<pcl::PointXYZ>::Ptr input, pcl::PointCloud<pcl::PointXYZ>::Ptr output, pcl::PointXYZ center, float x_min, float x_max, float y_min, float y_max,float z_min, float z_max)
{
	std::vector<int> indices;
	float cx = center.x;
	float cy = center.y;
	float cz = center.z;

	//std::cout << "x: " << center->points[center_indice].x << " y: " << center->points[center_indice].y << " z: " << center->points[center_indice].z << std::endl;

	for (int i = 0; i < input->size(); ++i)
	{	
		float dx = input->points[i].x - cx;
		float dy = input->points[i].y - cy;
		float dz = input->points[i].z - cz;

		//float length = sqrt(dx*dx + dy*dy + dz*dz);

		if ( dx < x_max and dx > x_min and dy < y_max and dy > y_min and dz < z_max and dz > z_min)
		{
			indices.push_back(i);
 		}
	}

	//std::cout << "Size of input: " << input->size() << ", Filtered indices: " << indices.size();
	pcl::copyPointCloud(*input, indices, *output);
	//std::cout << " Size of output: " << output->size() << std::endl;
}


void remove_plane(pcl::PointCloud<pcl::PointXYZ>::Ptr source, pcl::PointCloud<pcl::PointXYZ>::Ptr output)
{
	
	pcl::PointIndices::Ptr inliers (new pcl::PointIndices);

	pcl::SACSegmentation<pcl::PointXYZ> seg;
	seg.setOptimizeCoefficients(true);
	seg.setModelType(pcl::SACMODEL_PLANE);
	seg.setDistanceThreshold(0.0005);
	seg.setInputCloud(source);

	seg.segment(*inliers,*coefficients);
	if (inliers->indices.size () == 0)
	{
	   	PCL_ERROR ("Could not estimate a planar model for the given dataset.");
	}

	std::vector<int> outliers;

	for (int i = 0; i < source->size(); ++i)
	{
		if (!(std::find(inliers->indices.begin(),inliers->indices.end(), i) != inliers->indices.end())) outliers.push_back(i);
	}

	pcl::copyPointCloud(*source,outliers,*output);

}

typedef struct
{
	std::string name;
	int num;
	double time;
} cloud_t;

Eigen::Matrix4f transform(pcl::PointCloud<pcl::PointXYZ>::Ptr trajectory, double height, double time, double time_offset)
{
	double length = 0.25;
	double vel = 0.025; // Actually 50mm/s 
	double line_time = length / vel; 
	double x = 0.0;
	double y = 0.0;

	double real_time = time + time_offset;

	y = -real_time * vel;

	trajectory->height++;
	trajectory->resize(trajectory->width*trajectory->height);
	trajectory->points[trajectory->size()-1].x = x;
	trajectory->points[trajectory->size()-1].y = y;
	trajectory->points[trajectory->size()-1].z = height;

	Eigen::Matrix4f traj_transform = Eigen::Matrix4f::Identity();
	traj_transform (0,3) = x;
	traj_transform (1,3) = y;
	traj_transform (2,3) = height;

	/*Eigen::Matrix4f rot_z = Eigen::Matrix4f::Identity();
	rot_z (0,0) = cos(angle);
	rot_z (0,1) = sin(angle);
	rot_z (1,0) = -sin(angle);
	rot_z (1,1) = cos(angle);*/

	return traj_transform;
}

std::list<cloud_t> grab_file_names()
{
	std::list<cloud_t> result;
	std::ifstream file("filenames.txt");
	std::string line;

	while(std::getline(file,line))
	{
		cloud_t this_file;
		this_file.name = line;
		std::stringstream ss(line);
		std::string item;
		std::getline(ss, item, '_');
		std::getline(ss, item, '_');
		std::stringstream num(item);
		int y = 0;
		num >> y;
		this_file.num = y;
		std::getline(ss, item, 't');
		std::getline(ss, item, 'm');
		std::stringstream time(item);
		double x = 0.0;
		time >> x;
		this_file.time = x/1000;
		result.push_back(this_file);
	}

	file.close();

	return result;
}

void combine_clouds(pcl::PointCloud<pcl::PointXYZ>::Ptr world, pcl::PointCloud<pcl::PointXYZ>::Ptr now)
{
	
	*world += *now;
	pcl::PointCloud<pcl::PointXYZ> temp;

	pcl::VoxelGrid<pcl::PointXYZ> grid;
	grid.setInputCloud(world);
	grid.setLeafSize(0.0005f,0.0005f,0.0005f);
	grid.filter(temp);
	pcl::copyPointCloud(temp,*world);
}

std::vector<int> closest_points_to_cross_section(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr cs)
{
	std::vector<int> indices;
	for (int t_point = 0; t_point < cs->size(); t_point++)
	{
		float min_length = 1;
		int closest_indice = 0;
		float cx = cs->points[t_point].x;
		float cy = cs->points[t_point].y;
		float cz = cs->points[t_point].z;

		for (int c_point = 0; c_point < cloud->size(); c_point++)
		{
			float dx = cloud->points[c_point].x - cx;
			float dy = cloud->points[c_point].y - cy;
			float dz = cloud->points[c_point].z - cz;

			float length = sqrt(dx*dx + dy*dy + dz*dz);

			if (length < min_length)
			{
				min_length = length;
				closest_indice = c_point;
			}
		}
		if (!(std::find(indices.begin(),indices.end(), closest_indice) != indices.end())) indices.push_back(closest_indice);
	}
	return indices;
}

void calculate_real_trajectory(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr slice, pcl::PointCloud<pcl::PointXYZ>::Ptr real_trajectory, pcl::PointCloud<pcl::PointXYZ>::Ptr cross_section)
{
	pcl::PassThrough<pcl::PointXYZ> pass_y;
	pass_y.setFilterFieldName("y");
	pass_y.setFilterLimits(0.000, 0.0075);
	pass_y.setInputCloud(cloud);
	pass_y.filter(*slice);

	pcl::PassThrough<pcl::PointXYZ> pass_x;
	pass_y.setFilterFieldName("x");
	pass_y.setFilterLimits(-0.01, 0.01);
	pass_y.setInputCloud(slice);
	pass_y.filter(*slice);

	//auto closest_indices = closest_points_to_cross_section(slice, cross_section);
	//pcl::copyPointCloud(*slice, closest_indices, *slice);

	double ax = 0.0;
	double ay = 0.0;
	double az = 0.0;

	for (int i = 0; i < slice->size(); ++i)
	{
		ax += slice->points[i].x;
	}

	//real_trajectory->height++;
	//real_trajectory->resize(real_trajectory->width*real_trajectory->height);
	real_trajectory->points[real_trajectory->size()-1].x = ax/slice->size();
	real_trajectory->points[real_trajectory->size()-1].y = 0.0;
	real_trajectory->points[real_trajectory->size()-1].z = 0.0;
}

double calculate_road_width(pcl::PointCloud<pcl::PointXYZ>::Ptr slice)
{
	double y_thresh = 0.0001;
	double z_thresh = 0.0001; 

	double width = 0.0;

	for (int p = 0; p < slice->size(); p++)
	{
		double p_x = slice->points[p].x;
		double p_y = slice->points[p].y;
		double p_z = slice->points[p].z;

		for (int q = 0; q < slice->size(); q++)
		{
			double q_x = slice->points[q].x;
			double q_y = slice->points[q].y;
			double q_z = slice->points[q].z;

			double dx = p_x - q_x;	
			double dy = p_y - q_y;
			double dz = p_z - q_z;

			if (dy > y_thresh or dz > z_thresh)
			{
				continue;
			}

			if (dx > width)
			{
				width = dx;
			}
		}
	}
	return width;
}

float calculate_height(pcl::PointCloud<pcl::PointXYZRGB>::Ptr input, pcl::PointCloud<pcl::PointXYZRGB>::Ptr coloured, ext_edge_t edges, float min_y, float max_y)
{

	float z = 0.0;
	int z_iter = 0;

	float width = edges.p - edges.n;
	float neg_thres = edges.n + width/4.0;
	float pos_thres = edges.p - width/4.0;

	for (int p = 0; p < input->size(); ++p)
	{
		if (input->points[p].x >= neg_thres and input->points[p].x <= pos_thres and input->points[p].y < max_y and input->points[p].y > min_y)
		{
			z += input->points[p].z;
			z_iter++;

			coloured->points[p].r = 0;
			coloured->points[p].g = 0;
			coloured->points[p].b = 255;
		}
	}

	return z/static_cast<float>(z_iter);
}

int index_from_cords(int x, int y, int w)
{
	uint i = y * w + x;
	return i;
}

typedef struct {
	float gradient;
	float angle;
} edge_t;


float distance_between_points(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud)
{
	std::vector<float> dx;
	std::vector<float> dy;
	float x = 0.0;
	float y = 0.0;

	for (int i = 1; i < cloud->size(); ++i)
	{
		x = cloud->points[i].x - cloud->points[i-1].x;
		//std::cout << "dx: " << x << ", ";
		dx.push_back(x);
		y = cloud->points[i].y - cloud->points[i-1].y;
		//std::cout << "dy: " << y << std::endl;
		dy.push_back(y);
	}

	std::sort(dx.begin(),dx.end());
	std::sort(dy.begin(),dy.end());

	if (dx.size() % 2)
	{
		x = (dx[static_cast<int>(dx.size()/2)] + dx[static_cast<int>((dx.size()/2)) + 1])/2; 
	} else {
		x = dx[static_cast<int>(dx.size()/2)];
	}

	if (dy.size() % 2)
	{
		y = (dy[static_cast<int>(dy.size()/2)] + dy[static_cast<int>((dy.size()/2)) + 1])/2; 
	} else {
		y = dy[static_cast<int>(dy.size()/2)];
	}

	std::cout << "Median dx: " << x << ", Median dy: " << y << std::endl;

	//static float xy[2] = {x,y};

	return x;
}

int main()
{
	double time_offset = 0.0;

	pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>);
	pcl::PointCloud<pcl::PointXYZ>::Ptr filtered (new pcl::PointCloud<pcl::PointXYZ>);
	pcl::PointCloud<pcl::PointXYZ>::Ptr transformed (new pcl::PointCloud<pcl::PointXYZ>);
	pcl::PointCloud<pcl::PointXYZ>::Ptr zero (new pcl::PointCloud<pcl::PointXYZ>);
	pcl::PointCloud<pcl::PointXYZ>::Ptr zero_transformed (new pcl::PointCloud<pcl::PointXYZ>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr raw_cloud (new pcl::PointCloud<pcl::PointXYZRGB>);
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr smooth_cloud (new pcl::PointCloud<pcl::PointXYZRGB>);

	zero->width = 1;
	zero->height = 1;
	zero->resize(zero->width*zero->height);
	zero->points[zero->size()-1].x = 0.0;
	zero->points[zero->size()-1].y = 0.0;
	zero->points[zero->size()-1].z = 0.0;

	pcl::RadiusOutlierRemoval<pcl::PointXYZ> outrem;
	outrem.setRadiusSearch(0.001);
	outrem.setMinNeighborsInRadius(4);

	pcl::PassThrough<pcl::PointXYZ> pass_z;
	pass_z.setFilterFieldName("z");
	//pass_z.setFilterLimits(-0.001, top_of_road/1000);
	pcl::PassThrough<pcl::PointXYZ> pass_y;
	pass_y.setFilterFieldName("y");
	pass_y.setFilterLimits(0.0035, 0.025);

	int mode = 0;
	std::cout << "\nGaussian filtering? Yes/No(1,0)" << std::endl;
	std::cin >> mode;
	int kernel = 3;

	if(mode)
	{
		std::cout << "\nGaussian Filtering: Choose kernel size (3 or 5):" << std::endl;
		std::cin >> kernel ;
		std::cout << "You chose " << kernel << std::endl;	
	}
	float lower_z = 0.0;
	std::cout << "\nZ-threshold in mm: " << std::endl;
	std::cin >> lower_z;
	if(lower_z > 0.0)
	{
		lower_z = lower_z/1000;
	}
	
	std::cout << "Chosen lower z: " << lower_z << std::endl;

	boost::shared_ptr<pcl::visualization::PCLVisualizer> v (new pcl::visualization::PCLVisualizer ("3D Viewer"));

	std::list<cloud_t> file_names = grab_file_names();
	
	std::vector<float> all_errors;
	std::vector<float> all_y_pos;
	std::vector<float> all_road_widths;
	std::vector<float> all_heights;

	v->addCoordinateSystem(0.01);	
	v->setBackgroundColor (0.8,0.8,0.8);
	v->setSize(640, 480);
	v->setShowFPS(false);

	bool First = true;
	Eigen::Matrix4f calibration_transform;

	float min_y = 0.0025;
	float max_y = 0.0050;
	float vel = 0.025;
	float point_distance = 0.0005;
	float top_of_road = 0.01;

	for(std::list<cloud_t>::iterator it = file_names.begin(); it != file_names.end(); ++it)
	{
		std::cout << "Name: " <<  it->name << " Num: " << it->num << " Time: " << it->time << std::endl;
		float Y = it->time * vel;
		
		if (Y < 0.0475 and !First)
		{
			continue;
		}

		Y = Y - min_y;
		

		pcl::io::loadPCDFile<pcl::PointXYZ>(it->name, *cloud);
		raw_cloud->clear();

		if (First)
		{
			pcl::transformPointCloud(*cloud, *transformed, camera_transform());
			calibration_transform = calibrate_planes(transformed);
			pcl::transformPointCloud(*transformed, *transformed, calibration_transform);

			pcl::transformPointCloud(*zero, *zero_transformed, calibration_transform.inverse());
			pcl::transformPointCloud(*zero_transformed, *zero_transformed, camera_transform().inverse());

			point_distance = distance_between_points(transformed);
			top_of_road = calibration_transform(2,3);
			pass_z.setFilterLimits(-0.001, top_of_road-0.002);
			First = false;
			continue;
		}
		else
		{

			
			//std::cout << "z_top: " << z_top << std::endl;
			timer_start();
			range_filter_box_x_y_z(cloud, transformed, zero_transformed->points[0], -0.015, 0.015, -0.025, -min_y, -0.1, 0.1);
			std::cout << "range_filter_x_y_z: " << timer() << std::endl;


			timer_start();
			pcl::transformPointCloud(*transformed, *transformed, camera_transform());
			pcl::transformPointCloud(*transformed, *transformed, calibration_transform);
			std::cout << "transform: " << timer() << std::endl;

			timer_start();
			pass_z.setInputCloud(transformed);
			pass_z.filter(*transformed);

			pass_y.setInputCloud(transformed);
			pass_y.filter(*transformed);

			outrem.setInputCloud(transformed);
			outrem.filter(*transformed);
			std::cout << "filters: " << timer() << std::endl;


			pcl::copyPointCloud(*transformed, *raw_cloud);
		}

		if (mode)
		{
			timer_start();
			gaussian_blur(raw_cloud, smooth_cloud, kernel, point_distance);
			std::cout << "gaussian_blur: " << timer() << std::endl;
		} 
		else
		{
			pcl::copyPointCloud(*raw_cloud, *smooth_cloud);
		} 
		

		std::cout << "Y: " << Y << std::endl;
		
		v->addPointCloud<pcl::PointXYZRGB>(smooth_cloud,"world");
		v->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "world");
		v->spin();
		v->removePointCloud("world");

		timer_start();
		auto raw_gradients = calc_gradient(smooth_cloud, 2);	
		std::cout << "calc_gradient: " << timer() << std::endl;

		timer_start();
		auto edges = road_width_by_gradient(raw_gradients, smooth_cloud, min_y + 0.001, max_y+0.001, lower_z);
		std::cout << "find edges: " << timer() << std::endl;
	
		//auto smooth_gradients = calc_gradient(smooth_cloud, 2);

		float width = edges.p - edges.n;
		float error = edges.p - (width/2);
		std::cout << "p edge: " << edges.p << ", n edge: " << edges.n << ", width: " << width << std::endl;

		v->addPointCloud<pcl::PointXYZRGB>(smooth_cloud,"world");
		v->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 8, "world");
		v->spin();
		v->removePointCloud("world");		

		for (int i = 0; i < smooth_cloud->size(); ++i)
		{
			float x = smooth_cloud->points[i].x;
			float y = smooth_cloud->points[i].y;

			if (((x < edges.p + 0.0001 and x > edges.p - 0.0001) or (x < edges.n + 0.0001 and x > edges.n - 0.0001)) and (y < max_y))
			{
				smooth_cloud->points[i].r = 255; 
				smooth_cloud->points[i].g = 255;
				smooth_cloud->points[i].b = 255;
			} else {
				smooth_cloud->points[i].r = 0; 
				smooth_cloud->points[i].g = 0;
				smooth_cloud->points[i].b = 0;
			}
		}
		timer_start();
		auto height = calculate_height(raw_cloud, smooth_cloud, edges, min_y, max_y);
		std::cout << "calculate_height: " << timer() << std::endl;

		std::cout << "Height: " << height << std::endl;

		v->addPointCloud<pcl::PointXYZRGB>(smooth_cloud,"world");
		v->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 8, "world");
		v->spin();
		v->removePointCloud("world");

		all_errors.push_back(error);
		all_y_pos.push_back(Y);
		all_heights.push_back(height);
		all_road_widths.push_back(width);
	}

	std::ofstream error_text_file;
	std::stringstream ss;

	if (mode)
	{
		ss << "gaussian_values_" << kernel << ".txt";
	}
	else
	{
		ss << "unfiltered_values" << ".txt";
	}
	
	error_text_file.open(ss.str());

	double average_error = 0.0;
	int errors = all_errors.size();
	int iter = 0;

	error_text_file << "y pos\t ct error\t height\t roadwidth\n";

	for (auto &it : all_errors)
	{
		if(!std::isnan(all_y_pos[iter])) error_text_file << all_y_pos[iter]*1000 			<< "\t ";
		else error_text_file << 0.0 << "\t ";
		if(!std::isnan(all_errors[iter])) error_text_file << all_errors[iter]*1000 			<< "\t ";
		else error_text_file << 0.0 << "\t ";
		if(!std::isnan(all_road_widths[iter])) error_text_file << all_heights[iter]*1000 << "\t ";
		else error_text_file << 0.0 << "\t ";
		if(!std::isnan(all_heights[iter])) error_text_file << all_road_widths[iter]*1000;
		else error_text_file << 0.0;
		
		error_text_file << "\n";
		iter++;
	}	

	error_text_file.close();

	//ideal_cloud(raw_cloud);

	return 1;
}